"""
Package capturing all the readers available for parsing and processing data.
"""

from .coast import VtkCoastLineSource
from .mesh import VtkMeshReader
from .rivers import RiverReader
from .segment import VtkSegmentReader
from .topo import TopoReader

__all__ = [
    "RiverReader",
    "TopoReader",
    "VtkCoastLineSource",
    "VtkMeshReader",
    "VtkSegmentReader",
]



"""
Quad segment VTK reader from hdf5 file
"""

from pathlib import Path

import h5py
import numpy as np
from pyproj import Geod
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonCore import vtkFloatArray, vtkPoints
from vtkmodules.vtkCommonDataModel import vtkCellArray, vtkPolyData

from parsli.utils import earth

APPROXIMATE_SCALING = 111.0


def km_to_degrees_lon(km, lat):
    return km / (APPROXIMATE_SCALING * np.cos(np.deg2rad(lat)))


def km_to_degrees_lat(km):
    return km / APPROXIMATE_SCALING


def get_segment_bottom_lon_lat(lon1, lat1, lon2, lat2, locking_depth, dip):
    # Get segment azimuths
    azimuth = Geod(ellps="WGS84").inv(lon1, lat1, lon2, lat2)[0]

    # Get segment length
    length_km = locking_depth / np.tan(np.deg2rad(-1.0 * dip))

    # Get longitude and latitude spans
    delta_lon_km = length_km * np.cos(np.deg2rad(azimuth))
    delta_lat_km = -length_km * np.sin(np.deg2rad(azimuth))

    # Get average latitude
    avg_lat = (lat1 + lat2) / 2.0
    delta_lon = km_to_degrees_lon(delta_lon_km, avg_lat)
    delta_lat = km_to_degrees_lat(delta_lat_km)

    # Calculate approximate longitude and latitudes of lower vertices
    lon3 = lon1 + delta_lon
    lon4 = lon2 + delta_lon
    lat3 = lat1 + delta_lat
    lat4 = lat2 + delta_lat
    return lon3, lat3, lon4, lat4


class EarthLocation:
    __slots__ = ("lat", "lon")

    def __ilshift__(self, other):
        self.lat = other.lat
        self.lon = other.lon

    def flip(self):
        self.lon *= -1
        self.lat *= -1

    def interpolate_from(self, start_lon, start_lat, end_lon, end_lat, distance):
        self.lon, self.lat = earth.interpolate(
            start_lon, start_lat, end_lon, end_lat, distance
        )

    def __repr__(self):
        return f"Longitude: {self.lon}, Latitude: {self.lat}"


FIELD_COLS = {
    "strike_slip": 68,
    "dip_slip": 69,
    "tensile_slip": 70,
}
FIELD_NAMES = list(FIELD_COLS.keys())


class QuadCell:
    """
    Helper QuadCell for converting file data into actual coordinate for display.
    """

    __slots__ = (
        "dip",
        "end",
        "latitude_bnds",
        "locking_depth",
        "longitude_bnds",
        "normal",
        "point_a",
        "point_b",
        "start",
    )

    def __init__(self):
        self.start = EarthLocation()
        self.point_a = EarthLocation()
        self.point_b = EarthLocation()
        self.end = EarthLocation()
        self.normal = EarthLocation()

    def update(self, row):
        if row[34]:
            # skip cell if column 34 is true
            return False

        if row[0] >= row[2]:
            self.start.lon = row[0]
            self.start.lat = row[1]
            self.end.lon = row[2]
            self.end.lat = row[3]
        else:
            self.end.lon = row[0]
            self.end.lat = row[1]
            self.start.lon = row[2]
            self.start.lat = row[3]

        self.dip = row[4]
        self.locking_depth = row[14]

        lon3, lat3, lon4, lat4 = get_segment_bottom_lon_lat(
            self.start.lon,
            self.start.lat,
            self.end.lon,
            self.end.lat,
            self.locking_depth,
            self.dip,
        )

        self.point_a.lon = lon3
        self.point_a.lat = lat3
        self.point_b.lon = lon4
        self.point_b.lat = lat4

        return [(k, row[FIELD_COLS[k]]) for k in FIELD_NAMES]


class VtkSegmentReader(VTKPythonAlgorithmBase):
    """
    VTK Quad/Segment hdf5 reader
    """

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=0,
            nOutputPorts=1,
            outputType="vtkPolyData",
        )
        self._file_name = None
        self._proj_spherical = True
        self._field_names = []
        self._time_index = 0
        self._n_times = -1
        self._vertical_scale = 1
        self._valid = None

    @property
    def has_segments(self):
        """
        Check if the reader has segments.
        Some mesh files don't always contain segments.
        """
        if self._valid is None:
            self.Update()

        return self._valid

    @property
    def field_names(self):
        """
        Get the available field names.
        """
        if self.number_of_timesteps > 1:
            return self._field_names
        return FIELD_NAMES

    @property
    def file_name(self):
        """
        This property captures the file path to read.
        If set using an invalid path, a ValueError is raised.
        """
        return self._file_name

    @file_name.setter
    def file_name(self, path):
        self._file_name = Path(path)
        if not self._file_name.exists():
            msg = f"Invalid file path: {self._file_name.resolve()}"
            raise ValueError(msg)

        self._valid = None
        self._n_times = -1
        self._time_index = 0
        self.Modified()

    @property
    def spherical(self):
        """
        This property captures the projection system which can either be spherical or euclidean.
        When set to True (the default), the spherical projection will be used.
        """
        return self._proj_spherical

    @spherical.setter
    def spherical(self, value):
        if self._proj_spherical != value:
            self._proj_spherical = value
            self.Modified()

    @property
    def vertical_scale(self):
        """
        This property captures the vertical scale. The default value is 1.0.
        """
        return self._vertical_scale

    @vertical_scale.setter
    def vertical_scale(self, value):
        if self._vertical_scale != value:
            self._vertical_scale = value
            self.Modified()

    @property
    def time_index(self):
        """
        This property captures the current time_index.
        The value can only be within [0, number_of_timesteps] range.
        Setting outside values will be ignored.
        """
        return self._time_index

    @time_index.setter
    def time_index(self, v):
        if v == self._time_index:
            return

        if v == 0 or v < self.number_of_timesteps:
            self._time_index = v
            self.Modified()

    @property
    def number_of_timesteps(self):
        """
        This property captures the number of timesteps available in the current file.
        """
        if self._n_times < 0:
            with h5py.File(self._file_name, "r") as hdf:
                if "segments" in hdf:
                    segments = hdf["segments"]
                    self._field_names = list(segments.keys())
                    for field in segments:
                        n_times = len(segments[field].keys())
                        if n_times > self._n_times:
                            self._n_times = int(n_times)
                else:
                    self._n_times = 1

        return self._n_times

    def RequestData(self, _request, _inInfo, outInfo):
        if self._file_name is None or not self._file_name.exists():
            return 1

        # Read file and generate mesh
        output = self.GetOutputData(outInfo, 0)
        vtk_points = vtkPoints()
        vtk_points.SetDataTypeToDouble()
        vtk_polys = vtkCellArray()
        vtk_mesh = vtkPolyData()
        vtk_mesh.points = vtk_points
        vtk_mesh.polys = vtk_polys

        # Projection selection
        insert_pt = earth.insert_spherical if self.spherical else earth.insert_euclidian

        with h5py.File(self._file_name, "r") as hdf:
            self._valid = "segment" in hdf
            if not self._valid:
                return 1

            cell = QuadCell()
            h5_ds = hdf["segment"]
            data_size = h5_ds.shape

            # extract time dependent fields
            hdf_field_arrays = None
            if "segments" in hdf:
                hdf_field_arrays = {}
                segments = hdf["segments"]
                time_field_key = f"{self._time_index:012}"
                for field in segments:
                    hdf_field_arrays[field] = segments[field][time_field_key]

            # making a line for now (should move to 4 once quad)
            vtk_points.Allocate(data_size[0] * 2)
            vtk_polys.Allocate(data_size[0] * 5)

            # Create fields and attach to mesh
            vtk_field_arrays = {}
            for name in FIELD_NAMES:
                array = vtkFloatArray()
                array.SetName(name)
                array.Allocate(data_size[0])
                vtk_mesh.cell_data.AddArray(array)
                vtk_field_arrays[name] = array

            for row_idx, row in enumerate(h5_ds):
                if fields := cell.update(row):
                    vtk_polys.InsertNextCell(4)
                    vtk_polys.InsertCellPoint(
                        insert_pt(vtk_points, cell.start.lon, cell.start.lat, 0)
                    )
                    vtk_polys.InsertCellPoint(
                        insert_pt(
                            vtk_points,
                            cell.point_a.lon,
                            cell.point_a.lat,
                            cell.locking_depth * self._vertical_scale,
                        )
                    )
                    vtk_polys.InsertCellPoint(
                        insert_pt(
                            vtk_points,
                            cell.point_b.lon,
                            cell.point_b.lat,
                            cell.locking_depth * self._vertical_scale,
                        )
                    )
                    vtk_polys.InsertCellPoint(
                        insert_pt(vtk_points, cell.end.lon, cell.end.lat, 0)
                    )

                    # Add fields values
                    if hdf_field_arrays is None:
                        for k, v in fields:
                            vtk_field_arrays[k].InsertNextTuple1(v)
                    else:
                        for k, v in hdf_field_arrays.items():
                            vtk_field_arrays[k].InsertNextTuple1(v[row_idx])

        output.ShallowCopy(vtk_mesh)
        return 1



import base64
import json
import os
import queue
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import numpy as np
import vtkmodules.vtkRenderingOpenGL2  # noqa: F401
from vtkmodules.vtkCommonCore import vtkLookupTable, vtkUnsignedCharArray
from vtkmodules.vtkCommonDataModel import (
    vtkDataObject,
    vtkDataSetAttributes,
    vtkImageData,
)
from vtkmodules.vtkFiltersCore import (
    vtkArrayCalculator,
    vtkAssignAttribute,
    vtkCellDataToPointData,
    vtkThreshold,
)
from vtkmodules.vtkFiltersGeometry import vtkDataSetSurfaceFilter
from vtkmodules.vtkFiltersModeling import (
    vtkBandedPolyDataContourFilter,
    vtkLoopSubdivisionFilter,
)
from vtkmodules.vtkFiltersVerdict import vtkMeshQuality

# VTK factory initialization
from vtkmodules.vtkInteractionStyle import (
    vtkInteractorStyleSwitch,  # noqa: F401
    vtkInteractorStyleTerrain,
)
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkIOImage import vtkPNGWriter
from vtkmodules.vtkRenderingAnnotation import vtkAxesActor, vtkScalarBarActor
from vtkmodules.vtkRenderingCore import (
    vtkActor,
    vtkActor2D,
    vtkColorTransferFunction,
    vtkCompositePolyDataMapper,
    vtkPolyDataMapper,
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
    vtkTextMapper,
    vtkTextProperty,
    vtkWindowToImageFilter,
)

# Disable warning
vtkLoopSubdivisionFilter.GlobalWarningDisplayOff()


PRESETS = {
    item.get("Name"): item
    for item in json.loads(Path(__file__).with_name("presets.json").read_text())
}

LUTS = {}


def get_preset(preset_name: str) -> vtkColorTransferFunction:
    if preset_name in LUTS:
        return LUTS[preset_name]

    lut = LUTS.setdefault(preset_name, vtkColorTransferFunction())
    preset = PRESETS[preset_name]
    srgb = np.array(preset["RGBPoints"])
    color_space = preset["ColorSpace"]

    if color_space == "Diverging":
        lut.SetColorSpaceToDiverging()
    elif color_space == "HSV":
        lut.SetColorSpaceToHSV()
    elif color_space == "Lab":
        lut.SetColorSpaceToLab()
    elif color_space == "RGB":
        lut.SetColorSpaceToRGB()
    elif color_space == "CIELAB":
        lut.SetColorSpaceToLabCIEDE2000()

    if "NanColor" in preset:
        lut.SetNanColor(preset["NanColor"])

    # Always RGB points
    lut.RemoveAllPoints()
    for arr in np.split(srgb, len(srgb) / 4):
        lut.AddRGBPoint(arr[0], arr[1], arr[2], arr[3])

    return lut


def set_preset(lut: vtkLookupTable, preset_name: str, n_colors=255):
    colors = get_preset(preset_name)
    min, max = colors.GetRange()
    delta = max - min
    lut.SetNumberOfTableValues(n_colors)
    for i in range(n_colors):
        x = min + (delta * i / n_colors)
        rgb = colors.GetColor(x)
        lut.SetTableValue(i, *rgb)
    lut.Build()


def to_image(lut, samples=255):
    colorArray = vtkUnsignedCharArray()
    colorArray.SetNumberOfComponents(3)
    colorArray.SetNumberOfTuples(samples)

    dataRange = lut.GetRange()
    delta = (dataRange[1] - dataRange[0]) / float(samples)

    # Add the color array to an image data
    imgData = vtkImageData()
    imgData.SetDimensions(samples, 1, 1)
    imgData.GetPointData().SetScalars(colorArray)

    # Loop over all presets
    rgb = [0, 0, 0]
    for i in range(samples):
        lut.GetColor(dataRange[0] + float(i) * delta, rgb)
        r = round(rgb[0] * 255)
        g = round(rgb[1] * 255)
        b = round(rgb[2] * 255)
        colorArray.SetTuple3(i, r, g, b)

    writer = vtkPNGWriter()
    writer.WriteToMemoryOn()
    writer.SetInputData(imgData)
    writer.SetCompressionLevel(6)
    writer.Write()

    writer.GetResult()

    base64_img = base64.standard_b64encode(writer.GetResult()).decode("utf-8")
    return f"data:image/png;base64,{base64_img}"


def encode(writer, vtk_image, file_path, writer_queue):
    writer.file_name = file_path
    writer.input_data = vtk_image
    writer.Write()
    writer_queue.put(writer)


def update_range(lut: vtkColorTransferFunction, data_range):
    prev_min, prev_max = lut.GetRange()
    prev_delta = prev_max - prev_min

    if prev_delta < 0.001:
        return

    node = [0, 0, 0, 0, 0, 0]
    next_delta = data_range[1] - data_range[0]
    next_nodes = []

    for i in range(lut.GetSize()):
        lut.GetNodeValue(i, node)
        node[0] = next_delta * (node[0] - prev_min) / prev_delta + data_range[0]
        next_nodes.append(list(node))

    lut.RemoveAllPoints()
    for n in next_nodes:
        lut.AddRGBPoint(*n)


class SceneManager:
    def __init__(self, server):
        self.server = server

        self._lut = vtkLookupTable()
        self._lut_for_bar = vtkColorTransferFunction()
        set_preset(self._lut, "Fast")

        self.geometries = {}
        self.formulaFilters = []

        self.renderer = vtkRenderer(background=(1.0, 1.0, 1.0))
        self.interactor = vtkRenderWindowInteractor()
        self.render_window = vtkRenderWindow(off_screen_rendering=1)

        self.render_window.AddRenderer(self.renderer)
        self.interactor.SetRenderWindow(self.render_window)
        self.interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        self.style_terrain = vtkInteractorStyleTerrain()
        self.style_trackball = self.interactor.GetInteractorStyle()

        self.scalar_bar = vtkScalarBarActor()
        self.scalar_bar.SetOrientationToHorizontal()
        self.scalar_bar.SetLookupTable(self._lut_for_bar)
        self.scalar_bar.SetNumberOfLabels(3)
        self.scalar_bar.SetPosition(0.375, 0.01)
        self.scalar_bar.SetPosition2(0.25, 0.04)

        # labels
        self.scalar_bar.label_text_property.color = (0, 0, 0)
        self.scalar_bar.label_text_property.BoldOff()
        self.scalar_bar.label_text_property.ItalicOff()
        self.scalar_bar.SetTextPositionToPrecedeScalarBar()

        self.renderer.AddActor2D(self.scalar_bar)
        self.show_scalar_bar(False)

        # Time text
        text_property = vtkTextProperty()
        text_property.SetFontSize(16)
        text_property.SetJustificationToLeft()
        text_property.SetColor(0.2, 0.2, 0.2)

        self.time_mapper = vtkTextMapper(input="0.0", text_property=text_property)
        self.time_actor = vtkActor2D(mapper=self.time_mapper, position=(16, 16))
        self.renderer.AddActor(self.time_actor)

        camera = self.renderer.active_camera
        camera.position = (1, 0, 0)
        camera.focal_point = (0, 0, 0)
        camera.view_up = (0, 0, 1)

        self.interactor.Initialize()

        axes_actor = vtkAxesActor()
        self.widget = vtkOrientationMarkerWidget()
        self.widget.SetOrientationMarker(axes_actor)
        self.widget.SetInteractor(self.interactor)
        self.widget.SetViewport(0.85, 0, 1, 0.15)
        self.widget.EnabledOn()
        self.widget.InteractiveOff()

        self.window_image_filter = vtkWindowToImageFilter()
        self.window_image_filter.SetInput(self.render_window)
        self.window_image_filter.SetInputBufferTypeToRGB()

        print(f"Using {os.cpu_count()} thread for encoding")  # noqa: T201
        pool_size = os.cpu_count()
        self.encoder_pool = ThreadPoolExecutor(max_workers=pool_size)
        self.writer_queue = queue.Queue()

        # fill queue with writers
        self.writer_ext = ".png"
        for _ in range(pool_size):
            writer = vtkPNGWriter(compression_level=6)
            self.writer_queue.put(writer)

    @property
    def ctrl(self):
        return self.server.controller

    def set_time(self, time_value):
        self.time_mapper.input = f"t: {time_value}"

    def show_scalar_bar(self, show):
        self.scalar_bar.visibility = show

    def get_size(self):
        return self.render_window.GetSize()

    def set_size(self, width, height):
        self.render_window.SetSize(width, height)

    def screenshot(self):
        self.window_image_filter.Update()
        return self.window_image_filter.GetOutput()

    def write_screenshot(self, output_path):
        self.render_window.Render()
        self.window_image_filter.Modified()
        self.window_image_filter.Update()

        output = self.window_image_filter.output
        img = output.NewInstance()
        img.DeepCopy(output)

        writer = self.writer_queue.get()
        return self.encoder_pool.submit(
            encode, writer, img, f"{output_path}{self.writer_ext}", self.writer_queue
        )

    def __getitem__(self, key):
        return self.geometries.get(key)

    def reset_camera_to(self, bounds):
        self.renderer.ResetCamera(bounds)

    def reset_camera(self):
        self.renderer.ResetCamera()

    def focus_on(self, bounds):
        self.camera.position = (
            0.5 * (bounds[0] + bounds[1]),
            0.5 * (bounds[2] + bounds[3]),
            0.5 * (bounds[4] + bounds[5]),
        )
        self.reset_camera_to(bounds)

    def update_formula(self, formula_str):
        for filter in self.formulaFilters:
            filter.Update()
            ds = filter.GetInput()
            filter.RemoveScalarVariables()
            for field in ds.cell_data.keys():  # noqa: SIM118 (don't work on vtk new api)
                filter.AddScalarArrayName(field)
            filter.result_array_name = "formula"
            filter.function = formula_str

    def update_interaction_style(self, value):
        if value == "trackball":
            self.interactor.SetInteractorStyle(self.style_trackball)
        elif value == "terrain":
            camera = self.renderer.active_camera
            camera.view_up = (0, 0, 1)
            self.interactor.SetInteractorStyle(self.style_terrain)

    @property
    def camera(self):
        return self.renderer.active_camera

    @property
    def camera_state(self):
        return {
            "position": list(self.camera.position),
            "focal_point": list(self.camera.focal_point),
            "view_up": list(self.camera.view_up),
        }

    def update_camera(self, props):
        for k, v in props.items():
            setattr(self.renderer.active_camera, k, v)

    def update_view_up(self, view_up):
        self.renderer.active_camera.view_up = view_up

    def apply_zoom(self, scale):
        self.renderer.active_camera.Zoom(scale)

    def update_scalar_bar(self, color_preset, color_min, color_max):
        self._lut_for_bar.ShallowCopy(get_preset(color_preset))
        update_range(self._lut_for_bar, [color_min, color_max])

    def color_at(self, value, color_preset, color_min, color_max):
        color = [0, 0, 0]
        self.update_scalar_bar(color_preset, color_min, color_max)
        self._lut_for_bar.GetColor(value, color)
        return color

    @property
    def lut(self):
        return self._lut

    def add_geometry(self, name, source, composite=False, need_formula=False):
        item = {"name": name, "source": source, "composite": composite}

        if not composite:
            geometry = vtkDataSetSurfaceFilter(input_connection=source.output_port)
            mapper = vtkPolyDataMapper(
                input_connection=geometry.output_port,
                lookup_table=self.lut,
            )
            mapper.SetResolveCoincidentTopologyToOff()
            item["geometry"] = geometry
            item["mapper"] = mapper

            if need_formula:
                formula = vtkArrayCalculator(input_connection=source.output_port)
                formula.SetAttributeTypeToCellData()
                geometry.input_connection = formula.output_port
                self.formulaFilters.append(formula)
                item["formula"] = formula
        else:
            mapper = vtkCompositePolyDataMapper(
                input_connection=source.output_port,
                lookup_table=self.lut,
                interpolate_scalars_before_mapping=1,
            )

            if need_formula:
                formula = vtkArrayCalculator(input_connection=source.output_port)
                formula.SetAttributeTypeToCellData()
                mapper.input_connection = formula.output_port
                self.formulaFilters.append(formula)
                item["formula"] = formula

            item["mapper"] = mapper
            mapper.SetResolveCoincidentTopologyToOff()

        actor = vtkActor(mapper=mapper)
        item["actor"] = actor

        self.geometries[name] = item

        self.renderer.AddActor(actor)
        self.renderer.ResetCamera()
        self.render_window.Render()

        self.ctrl.view_update()

        # Gather actors/mappers
        item["actors"] = [
            actor,
        ]
        item["mappers"] = [
            item["mapper"],
        ]

        return item

    def add_geometry_with_contour(
        self, name, source, composite=False, need_formula=False, field_name="dip_slip"
    ):
        # pipeline filters
        quality = vtkMeshQuality()
        quality.SetTriangleQualityMeasureToEdgeRatio()
        threshold = vtkThreshold(
            threshold_function=vtkThreshold.THRESHOLD_LOWER,
            lower_threshold=3.99,  # magic number for vtkLoopSubdivisionFilter
        )
        threshold.SetInputArrayToProcess(
            0, 0, 0, vtkDataObject.FIELD_ASSOCIATION_CELLS, "Quality"
        )
        geometry = vtkDataSetSurfaceFilter()
        cell2point = vtkCellDataToPointData()
        refine = vtkLoopSubdivisionFilter(
            number_of_subdivisions=1
        )  # Adjust subdivision quality
        assign = vtkAssignAttribute()
        assign.Assign(
            field_name,
            vtkDataSetAttributes.SCALARS,
            vtkDataObject.FIELD_ASSOCIATION_POINTS,
        )
        bands = vtkBandedPolyDataContourFilter(generate_contour_edges=1)

        # connect pipeline
        (
            source
            >> quality
            >> threshold
            >> geometry
            >> cell2point
            >> refine
            >> assign
            >> bands
        )
        for_surface = bands

        # bands.Update()
        # print(bands.GetOutputDataObject(0))

        item = {
            "name": name,
            "source": source,
            "quality": quality,
            "threshold": threshold,
            "geometry": geometry,
            "composite": composite,
            "cell2point": cell2point,
            "refine": refine,
            "assign": assign,
            "bands": bands,
        }

        if need_formula:
            formula = vtkArrayCalculator()
            formula.SetAttributeTypeToCellData()
            self.formulaFilters.append(formula)
            (
                source
                >> formula
                >> quality
                >> threshold
                >> geometry
                >> cell2point
                >> refine
                >> assign
                >> bands
            )
            item["formula"] = formula

        if not composite:
            # surface
            mapper = vtkPolyDataMapper(
                input_connection=for_surface.output_port,
                lookup_table=self.lut,
                interpolate_scalars_before_mapping=1,
            )
            mapper.SelectColorArray("Scalars")
            mapper.SetResolveCoincidentTopologyToOff()
            item["mapper"] = mapper
            # lines
            mapper_lines = vtkPolyDataMapper(
                input_connection=bands.GetOutputPort(1),
            )
            mapper_lines.SetResolveCoincidentTopologyToPolygonOffset()
            item["mapper_lines"] = mapper_lines
        else:
            # surface
            mapper = vtkCompositePolyDataMapper(
                input_connection=for_surface.output_port,
                lookup_table=self.lut,
                interpolate_scalars_before_mapping=1,
            )
            mapper.SelectColorArray("Scalars")
            mapper.SetResolveCoincidentTopologyToOff()
            item["mapper"] = mapper
            # lines
            mapper_lines = vtkCompositePolyDataMapper(
                input_connection=bands.GetOutputPort(1),
                scalar_visibility=0,
            )
            mapper_lines.SetResolveCoincidentTopologyToPolygonOffset()
            item["mapper_lines"] = mapper_lines

        # Surface actor
        actor = vtkActor(mapper=mapper)
        item["actor"] = actor

        # Lines actor
        actor_lines = vtkActor(mapper=mapper_lines)
        actor_lines.property.color = (0, 0, 0)
        actor_lines.property.line_width = 2
        # actor_lines.property.render_line_as_tube = 1
        item["actor_lines"] = actor_lines

        self.geometries[name] = item

        self.renderer.AddActor(actor)
        self.renderer.AddActor(actor_lines)
        self.renderer.ResetCamera()
        self.render_window.Render()

        self.ctrl.view_update()

        # Gather actors/mappers
        item["actors"] = [
            item["actor"],
            item["actor_lines"],
        ]
        item["mappers"] = [
            item["mapper"],
            item["mapper_lines"],
        ]

        return item

    def __contains__(self, name):
        return name in self.geometries



from __future__ import annotations

import logging
from pathlib import Path

import h5py
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import vtkStructuredGrid
from vtkmodules.vtkCommonExecutionModel import vtkStreamingDemandDrivenPipeline

from parsli.utils import earth

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class TopoReader(VTKPythonAlgorithmBase):
    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=0,
            nOutputPorts=1,
            outputType="vtkStructuredGrid",
        )
        self._file_name = None
        self._proj_spherical = True
        self._longitude_bnd = [0, 360]
        self._latitude_bnd = [-90, 90]
        self._vertical_scale = 1
        self._max_depth = 0

    @property
    def file_name(self):
        return self._file_name

    @file_name.setter
    def file_name(self, path):
        self._file_name = Path(path)
        if not self._file_name.exists():
            msg = f"Invalid file path: {self._file_name.resolve()}"
            raise ValueError(msg)

        self.Modified()

    @property
    def spherical(self):
        return self._proj_spherical

    @spherical.setter
    def spherical(self, value):
        if self._proj_spherical != value:
            self._proj_spherical = value
            self.Modified()

    @property
    def vertical_scale(self):
        return self._vertical_scale

    @vertical_scale.setter
    def vertical_scale(self, value):
        if self._vertical_scale != value:
            self._vertical_scale = value
            self.Modified()

    @property
    def maximum_depth(self):
        return self._max_depth

    @property
    def longitude_bounds(self):
        return self._longitude_bnd

    @property
    def latitude_bounds(self):
        return self._latitude_bnd

    def _expand_bounds(self, longitude, latitude, depth):
        self._longitude_bnd[0] = min(longitude, self._longitude_bnd[0])
        self._longitude_bnd[1] = max(longitude, self._longitude_bnd[1])

        self._latitude_bnd[0] = min(latitude, self._latitude_bnd[0])
        self._latitude_bnd[1] = max(latitude, self._latitude_bnd[1])

        self._max_depth = max(depth, self._max_depth)

    def RequestInformation(self, request, inInfo, outInfo):
        with h5py.File(self._file_name, "r") as hdf:
            dims = hdf["topo/elevation"].shape
            whole_extent = (0, dims[1] - 1, 0, dims[0] - 1, 0, 0)
            outInfo.GetInformationObject(0).Set(
                vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(),
                *whole_extent,
            )

        return 1

    def RequestData(self, _request, _inInfo, outInfo):
        if self._file_name is None or not self._file_name.exists():
            return 1

        # Reset bounds
        self._longitude_bnd = [360, 0]
        self._latitude_bnd = [90, -90]
        self._max_depth = 0

        # Read file and generate mesh
        output = self.GetOutputData(outInfo, 0)
        topo = vtkStructuredGrid()

        # Projection selection
        insert_pt = earth.insert_spherical if self.spherical else earth.insert_euclidian

        with h5py.File(self._file_name, "r") as hdf:
            lon_lat_bounds = hdf["topo/bounds"]
            elevations = hdf["topo/elevation"]
            dims = elevations.shape

            # Extract data bounds
            min_lon = lon_lat_bounds[0][0]
            max_lon = lon_lat_bounds[0][1]
            delta_lon = max_lon - min_lon
            min_lat = lon_lat_bounds[1][0]
            max_lat = lon_lat_bounds[1][1]
            delta_lat = max_lat - min_lat

            # Generate points
            vtk_points = vtkPoints()
            vtk_points.SetDataTypeToDouble()
            vtk_points.Allocate(dims[0] * dims[1])

            for j in range(dims[0]):
                lat = min_lat + delta_lat * j / (dims[0] - 1)
                for i in range(dims[1]):
                    lon = min_lon + delta_lon * i / (dims[1] - 1)
                    elevation = elevations[j][i]

                    self._expand_bounds(lon, lat, elevation * self._vertical_scale)
                    insert_pt(vtk_points, lon, lat, elevation * self._vertical_scale)

            topo.SetDimensions(dims[1], dims[0], 1)
            topo.points = vtk_points

        output.ShallowCopy(topo)
        return 1



"""
Topography Tools that will generate a topo.hdf5 file using some
wavy shape.

Can be ran using the following command line:

   python -m parsli.tools.topo

"""

import h5py
import numpy as np

output_file = "./data/topo.hdf5"

min_lon = 231.8
max_lon = 238.2
min_lat = 39.4
max_lat = 51.1

lon_step_size = 1 / 3
lat_step_size = 1 / 3

# Each river will be a sinusoidal along the grids.
# Set parameters for each sinusoidal.
river_params = [
    # amplitude, period, slope, and intercept
    (0.1, 1, 1, 0),
    (0.2, 0.5, 0.5, -0.5),
    (0.3, 2, 0, 0.5),
]
num_river_coords = 100
river_step_size = 1 / (num_river_coords - 1)

# Compute the elevation grid
lon_vals = np.arange(min_lon, max_lon + lon_step_size, lon_step_size)
lat_vals = np.arange(min_lat, max_lat + lat_step_size, lat_step_size)
lon_grid, lat_grid = np.meshgrid(lon_vals, lat_vals)
elevation_grid = 15 * (np.cos(lon_grid * 2) + np.sin(lat_grid * 2))

# Compute the rivers
rivers = {}
river_idx = 1
for A, B, m, c in river_params:
    # Compute the river's fractional coordinates
    x_coords = np.arange(0, 1 + river_step_size, river_step_size)
    y_coords = A * np.sin(B * x_coords) + m * x_coords + c

    # Remove any out-of-bounds coordinates
    to_keep = ~((y_coords < 0) | (y_coords > 1))
    x_coords = x_coords[to_keep]
    y_coords = y_coords[to_keep]

    # Convert these to latitude and longitude
    # (0, 0) is at (min_lon, max_lat), and (1, 1) is at (max_lon, min_lat)
    lon_coords = x_coords * (max_lon - min_lon) + min_lon
    lat_coords = (1 - y_coords) * (max_lat - min_lat) + min_lat
    elevation = 15 * (np.cos(lon_coords * 2) + np.sin(lat_coords * 2))
    rivers[river_idx] = np.vstack((lon_coords, lat_coords, elevation)).T
    river_idx += 1

# Now write everything to the HDF5 file
with h5py.File(output_file, "w") as f:
    f["topo/bounds"] = [[min_lon, max_lon], [min_lat, max_lat]]
    f["topo/elevation"] = elevation_grid
    for idx, data in rivers.items():
        f[f"lines/river{idx}"] = data



"""
Image Player application

Usage:

    # Using module name
    python -m parsli.player.app --data /path/to/parsli/export1/ /path/to/parsli/export2/ ...

    # Using script alias
    parsli-player --data /path/to/parsli/export1/ /path/to/parsli/export2/ ...

"""

import asyncio
from pathlib import Path

from trame.app import TrameApp, asynchronous
from trame.decorators import change
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import vuetify3 as v3
from trame.widgets.trame import MouseTrap

LAYOUTS = [
    {"title": 1, "value": 12},
    {"title": 2, "value": 6},
    {"title": 3, "value": 4},
    {"title": 4, "value": 3},
]


class ParsliPlayer(TrameApp):
    def __init__(self, server=None):
        super().__init__(server, client_type="vue3")
        self.server.cli.add_argument(
            "--data", nargs="+", help="List all export directory you want to load"
        )
        args, _ = self.server.cli.parse_known_args()

        # Serve datasets over http
        http_serve = {}
        n_timesteps = []
        for idx, data_path in enumerate(args.data, start=1):
            ds_path = Path(data_path).resolve()
            n_timesteps.append(len(list(ds_path.glob("*.png"))))
            http_serve[f"{idx}"] = str(ds_path)
        self.server.enable_module({"serve": http_serve})

        # Extract dataset info
        self.state.trame__title = "Parsli Player"
        self.state.nb_datasets = len(args.data)
        self.state.time_max = n_timesteps[0] - 1
        self.state.time_index = 0
        assert sum(n_timesteps) == n_timesteps[0] * self.state.nb_datasets

        # Build UI
        with VAppLayout(self.server, full_height=True) as self.ui:
            with MouseTrap(
                next="time_index < time_max && time_index++",
                prev="time_index && time_index--",
                start="time_index = 0",
                end="time_index = time_max",
                togglePlay="playing = !playing",
            ) as mt:
                mt.bind(["right"], "next")
                mt.bind(["left"], "prev")
                mt.bind(["space"], "togglePlay")
                mt.bind(["home"], "start")
                mt.bind(["end"], "end")
            with v3.VLayout():
                with v3.VMain():
                    with v3.VContainer(classes="h-100 pa-0 ma-0", fluid=True):
                        with v3.VRow(no_gutters=True):
                            with v3.VCol(
                                cols=("nb_cols",),
                                v_for="ds_idx in nb_datasets",
                                key="ds_idx",
                            ):
                                v3.VImg(
                                    src=(
                                        "`/${ds_idx}/${('000000000000' + time_index).slice(-12)}.png`",
                                    ),
                                )
                with v3.VFooter(app=True):
                    v3.VSelect(
                        v_model=("nb_cols", 6),
                        items=("layouts", LAYOUTS),
                        density="compact",
                        hide_details=True,
                        variant="solo",
                        flat=True,
                        style="max-width: 5rem;",
                    )
                    v3.VSlider(
                        v_model=("time_index", 0),
                        max=("time_max", 1),
                        min=0,
                        step=1,
                        density="compact",
                        hide_details=True,
                        classes="mx-2",
                    )
                    v3.VCheckbox(
                        v_model=("playing", False),
                        true_icon="mdi-stop",
                        false_icon="mdi-play",
                        density="compact",
                        hide_details=True,
                        classes="mx-2",
                    )
                    v3.VLabel(
                        "{{ time_index + 1}}/{{ time_max + 1}}",
                        style="width: 100px;",
                        classes="text-right",
                    )

    @property
    def state(self):
        return self.server.state

    @change("playing")
    def on_play(self, playing, **_):
        if playing:
            asynchronous.create_task(self._playing())

    async def _playing(self):
        while self.state.playing:
            with self.state:
                if self.state.time_index < self.state.time_max:
                    self.state.time_index += 1
                else:
                    self.state.playing = False
            await asyncio.sleep(0.1)


def main():
    app = ParsliPlayer()
    app.server.start()


if __name__ == "__main__":
    main()



METADATA_STATE_KEYS = [
    # Projection
    "spherical",
    # Cropping bounds
    "longitude_bnds",
    "latitude_bnds",
    # Vertical scaling
    "vertical_scaling",
    "vertical_scaling_topo",
    # Grid Box
    "nb_grid_line_per_degree",
    "show_grid",
    # Earth visibility
    "show_earth_core",
    # Coast line regions
    "coast_active_regions",
    # Do we have topo
    "topo_ui",
    # Visibility of meshes
    "show_surface",
    "show_segment",
    "show_terrain",
    "show_rivers",
    # Shadow of meshes
    "light_surface",
    "light_segment",
    "light_terrain",
    "light_rivers",
    # Opacity of meshes
    "surface_opacity",
    "segment_opacity",
    "terrain_opacity",
    "rivers_opacity",
    # Coloring information
    "color_by",
    "color_min",
    "color_max",
    "color_preset",
    # Processing pipeline
    "subdivide",
    "nb_contours",
    "use_formula",
    "formula",
]



"""
Spherical Axis Grid Box
"""

import math

from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonCore import vtkPoints, vtkTypeFloat32Array
from vtkmodules.vtkCommonDataModel import (
    vtkCellArray,
    vtkPlanes,
    vtkPolyData,
)

from parsli.utils import earth


class VtkLatLonBound(VTKPythonAlgorithmBase):
    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=0,
            nOutputPorts=1,
            outputType="vtkPolyData",
        )
        self._file_name = None
        self._proj_spherical = True
        self._longitude_bnd = [0, 360]
        self._latitude_bnd = [-90, 90]
        self._sampling_per_degree = 5
        self._depth = 100
        self._cut_planes = vtkPlanes()
        self._cut_planes_origin = vtkPoints()
        self._cut_planes_normal = vtkTypeFloat32Array()
        self._cut_planes_normal.SetNumberOfComponents(3)
        self._cut_planes.SetPoints(self._cut_planes_origin)
        self._cut_planes.SetNormals(self._cut_planes_normal)
        self._n_gridline_per_degree = 1
        self._show_grid_lines = True

    @property
    def cut_planes(self):
        return self._cut_planes

    @property
    def spherical(self):
        return self._proj_spherical

    @spherical.setter
    def spherical(self, value):
        if self._proj_spherical != value:
            self._proj_spherical = value
            self.Modified()

    @property
    def grid_lines(self):
        return self._show_grid_lines

    @grid_lines.setter
    def grid_lines(self, value):
        if self._show_grid_lines != value:
            self._show_grid_lines = value
            self.Modified()

    @property
    def grid_lines_per_degree(self):
        return self._show_grid_lines

    @grid_lines_per_degree.setter
    def grid_lines_per_degree(self, value):
        if self._n_gridline_per_degree != value:
            self._n_gridline_per_degree = value
            self.Modified()

    @property
    def longitude_bnds(self):
        return self._longitude_bnd

    @longitude_bnds.setter
    def longitude_bnds(self, lon_bnd):
        if self._longitude_bnd != lon_bnd:
            self._longitude_bnd = lon_bnd
            self.Modified()

    @property
    def latitude_bnds(self):
        return self._latitude_bnd

    @latitude_bnds.setter
    def latitude_bnds(self, lat_bnd):
        if self._latitude_bnd != lat_bnd:
            self._latitude_bnd = lat_bnd
            self.Modified()

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, new_depth):
        if self._depth != new_depth:
            self._depth = new_depth
            self.Modified()

    @property
    def valid(self):
        if self._proj_spherical:
            delta_lon = self._longitude_bnd[1] - self._longitude_bnd[0]
            delta_lat = self._latitude_bnd[1] - self._latitude_bnd[0]
            if delta_lon > 180 or delta_lat > 90:
                return False

        return True

    def RequestData(self, _request, _inInfo, outInfo):
        self.cut_planes.Modified()

        if not self.valid:
            self._cut_planes_origin.SetNumberOfPoints(0)
            self._cut_planes_normal.SetNumberOfTuples(0)
            return 1

        # Read file and generate mesh
        output = self.GetOutputData(outInfo, 0)

        vtk_mesh = vtkPolyData()
        vtk_points = vtkPoints()
        vtk_points.SetDataTypeToDouble()
        vtk_mesh.points = vtk_points
        vtk_lines = vtkCellArray()
        vtk_mesh.lines = vtk_lines

        # Projection selection
        if self.spherical:
            insert_pt = earth.insert_spherical
            delta_lon = self._longitude_bnd[1] - self._longitude_bnd[0]
            delta_lat = self._latitude_bnd[1] - self._latitude_bnd[0]
            mid_lon = 0.5 * (self._longitude_bnd[0] + self._longitude_bnd[1])

            n_lon_pts = int(delta_lon * self._sampling_per_degree + 0.5)
            n_lat_pts = int(delta_lat * self._sampling_per_degree + 0.5)

            if self.grid_lines:
                if self.grid_lines_per_degree < 0:
                    # 1 line per n degree
                    nb_points = n_lon_pts * 4 + n_lat_pts * 4
                    nb_lines = (1 + n_lon_pts) * 4 + (1 + n_lat_pts) * 4 + 8

                    modulo = abs(self._n_gridline_per_degree)

                    # vertical lines
                    lon = math.floor(self.longitude_bnds[0])
                    while lon < self.longitude_bnds[1]:
                        if lon < self.longitude_bnds[0]:
                            lon += 1
                            continue

                        if lon % modulo != 0:
                            lon += 1
                            continue

                        nb_points += n_lat_pts
                        nb_lines += 1 + n_lat_pts
                        lon += modulo

                    # horizontal lines
                    lat = math.floor(self._latitude_bnd[0])
                    while lat < self._latitude_bnd[1]:
                        if lat < self._latitude_bnd[0]:
                            lat += 1
                            continue

                        if lat % modulo != 0:
                            lat += 1
                            continue

                        nb_points += n_lon_pts
                        nb_lines += 1 + n_lon_pts
                        lat += modulo

                    vtk_points.Allocate(nb_points)
                    nb_lines.Allocate(nb_lines)
                else:
                    # n lines per 1 degree
                    n_vlines = int(delta_lon * self.grid_lines_per_degree)
                    n_hlines = int(delta_lat * self.grid_lines_per_degree)
                    vtk_points.Allocate(
                        n_lon_pts * 4  # bbox horizontal lines pts
                        + n_lat_pts * 4  # bbox vertical lines pts
                        + n_lon_pts * n_hlines  # grid horizontal pts
                        + n_lat_pts * n_vlines  # grid vertical pts
                    )
                    vtk_lines.Allocate(
                        (1 + n_lon_pts) * 4  # horizontal bbox lines
                        + (1 + n_lat_pts) * 4  # vertical bbox lines
                        + 8  # depth lines
                        + n_vlines * n_lat_pts  # vertical grid lines
                        + n_hlines * n_lon_pts  # horizontal grid lines
                    )

            else:
                vtk_points.Allocate(n_lon_pts * 4 + n_lat_pts * 4)
                vtk_lines.Allocate((1 + n_lon_pts) * 4 + (1 + n_lat_pts) * 4 + 8)

            # Compute cut planes
            self._cut_planes_origin.SetNumberOfPoints(4)
            self._cut_planes_normal.SetNumberOfTuples(4)

            # top
            lon = mid_lon
            lat = self._latitude_bnd[1]
            center = earth.to_spherical(lon, lat, 0)
            normal = earth.to_normal(center, (0, 0, -1))
            self._cut_planes_origin.SetPoint(0, center)
            self._cut_planes_normal.SetTuple3(0, *normal)

            # bottom
            lon = mid_lon
            lat = self._latitude_bnd[0]
            center = earth.to_spherical(lon, lat, 0)
            self._cut_planes_origin.SetPoint(1, center)
            self._cut_planes_normal.SetTuple3(1, *earth.to_normal(center, (0, 0, 1)))

            # right
            self._cut_planes_origin.SetPoint(2, 0, 0, 0)
            self._cut_planes_normal.SetTuple3(
                2, *earth.left_direction(self._longitude_bnd[1])
            )

            # left
            self._cut_planes_origin.SetPoint(3, 0, 0, 0)
            self._cut_planes_normal.SetTuple3(
                3, *earth.right_direction(self._longitude_bnd[0])
            )

            # Generate points and cells
            for depth in [self._depth, 0]:
                # Start bottom left
                lon = self._longitude_bnd[0]
                lat = self._latitude_bnd[0]

                # Bottom
                vtk_lines.InsertNextCell(n_lon_pts)
                for lon_idx in range(n_lon_pts):
                    lon = self._longitude_bnd[0] + lon_idx * delta_lon / (n_lon_pts - 1)
                    vtk_lines.InsertCellPoint(insert_pt(vtk_points, lon, lat, depth))

                # Right
                vtk_lines.InsertNextCell(n_lat_pts)
                for lat_idx in range(n_lat_pts):
                    lat = self._latitude_bnd[0] + lat_idx * delta_lat / (n_lat_pts - 1)
                    vtk_lines.InsertCellPoint(insert_pt(vtk_points, lon, lat, depth))

                # Top
                vtk_lines.InsertNextCell(n_lon_pts)
                for lon_idx in range(n_lon_pts)[::-1]:
                    lon = self._longitude_bnd[0] + lon_idx * delta_lon / (n_lon_pts - 1)
                    vtk_lines.InsertCellPoint(insert_pt(vtk_points, lon, lat, depth))

                # Left
                vtk_lines.InsertNextCell(n_lat_pts)
                for lat_idx in range(n_lat_pts)[::-1]:
                    lat = self._latitude_bnd[0] + lat_idx * delta_lat / (n_lat_pts - 1)
                    vtk_lines.InsertCellPoint(insert_pt(vtk_points, lon, lat, depth))

            # Generate depth lines
            next_layer_offset = 2 * n_lon_pts + 2 * n_lat_pts
            steps = [0, n_lon_pts, n_lat_pts, n_lon_pts]
            offset = 0
            for step in steps:
                offset += step
                vtk_lines.InsertNextCell(2)
                vtk_lines.InsertCellPoint(offset)
                vtk_lines.InsertCellPoint(offset + next_layer_offset)

            # Generate grid lines
            if self.grid_lines:
                if self._n_gridline_per_degree < 0:
                    # number of degree between lines
                    modulo = abs(self._n_gridline_per_degree)

                    # vertical lines
                    lon = math.floor(self.longitude_bnds[0])
                    while lon < self.longitude_bnds[1]:
                        if lon < self.longitude_bnds[0]:
                            lon += 1
                            continue

                        if lon % modulo != 0:
                            lon += 1
                            continue

                        vtk_lines.InsertNextCell(n_lat_pts)
                        for lat_idx in range(n_lat_pts):
                            lat = self._latitude_bnd[0] + lat_idx * delta_lat / (
                                n_lat_pts - 1
                            )
                            vtk_lines.InsertCellPoint(
                                insert_pt(vtk_points, lon, lat, self._depth)
                            )

                        lon += modulo

                    # horizontal lines
                    lat = math.floor(self._latitude_bnd[0])
                    while lat < self._latitude_bnd[1]:
                        if lat < self._latitude_bnd[0]:
                            lat += 1
                            continue

                        if lat % modulo != 0:
                            lat += 1
                            continue

                        vtk_lines.InsertNextCell(n_lon_pts)
                        for lon_idx in range(n_lon_pts):
                            lon = self._longitude_bnd[0] + lon_idx * delta_lon / (
                                n_lon_pts - 1
                            )
                            vtk_lines.InsertCellPoint(
                                insert_pt(vtk_points, lon, lat, self._depth)
                            )

                        lat += modulo
                else:
                    # vertical lines
                    step = 1 / self._n_gridline_per_degree
                    lon = math.floor(self.longitude_bnds[0])
                    while lon < self.longitude_bnds[1]:
                        if lon < self.longitude_bnds[0]:
                            lon += step
                            continue

                        vtk_lines.InsertNextCell(n_lat_pts)
                        for lat_idx in range(n_lat_pts):
                            lat = self._latitude_bnd[0] + lat_idx * delta_lat / (
                                n_lat_pts - 1
                            )
                            vtk_lines.InsertCellPoint(
                                insert_pt(vtk_points, lon, lat, self._depth)
                            )

                        lon += step

                    # horizontal lines
                    step = 1 / self._n_gridline_per_degree
                    lat = math.floor(self._latitude_bnd[0])
                    while lat < self._latitude_bnd[1]:
                        if lat < self._latitude_bnd[0]:
                            lat += step
                            continue

                        vtk_lines.InsertNextCell(n_lon_pts)
                        for lon_idx in range(n_lon_pts):
                            lon = self._longitude_bnd[0] + lon_idx * delta_lon / (
                                n_lon_pts - 1
                            )
                            vtk_lines.InsertCellPoint(
                                insert_pt(vtk_points, lon, lat, self._depth)
                            )

                        lat += step

        else:
            vtk_points.Allocate(8)

            vtk_points.InsertNextPoint(
                self._longitude_bnd[0], self._latitude_bnd[0], -self._depth
            )
            vtk_points.InsertNextPoint(
                self._longitude_bnd[1], self._latitude_bnd[0], -self._depth
            )
            vtk_points.InsertNextPoint(
                self._longitude_bnd[1], self._latitude_bnd[1], -self._depth
            )
            vtk_points.InsertNextPoint(
                self._longitude_bnd[0], self._latitude_bnd[1], -self._depth
            )

            vtk_points.InsertNextPoint(self._longitude_bnd[0], self._latitude_bnd[0], 0)
            vtk_points.InsertNextPoint(self._longitude_bnd[1], self._latitude_bnd[0], 0)
            vtk_points.InsertNextPoint(self._longitude_bnd[1], self._latitude_bnd[1], 0)
            vtk_points.InsertNextPoint(self._longitude_bnd[0], self._latitude_bnd[1], 0)

            vtk_lines.Allocate(5 + 5 + 2 * 4)

            # Bottom
            vtk_lines.InsertNextCell(5)
            vtk_lines.InsertCellPoint(0)
            vtk_lines.InsertCellPoint(1)
            vtk_lines.InsertCellPoint(2)
            vtk_lines.InsertCellPoint(3)
            vtk_lines.InsertCellPoint(0)

            # Top
            vtk_lines.InsertNextCell(5)
            vtk_lines.InsertCellPoint(4)
            vtk_lines.InsertCellPoint(5)
            vtk_lines.InsertCellPoint(6)
            vtk_lines.InsertCellPoint(7)
            vtk_lines.InsertCellPoint(4)

            # Edges
            vtk_lines.InsertNextCell(2)
            vtk_lines.InsertCellPoint(0)
            vtk_lines.InsertCellPoint(4)

            vtk_lines.InsertNextCell(2)
            vtk_lines.InsertCellPoint(1)
            vtk_lines.InsertCellPoint(5)

            vtk_lines.InsertNextCell(2)
            vtk_lines.InsertCellPoint(2)
            vtk_lines.InsertCellPoint(6)

            vtk_lines.InsertNextCell(2)
            vtk_lines.InsertCellPoint(3)
            vtk_lines.InsertCellPoint(7)

            # Compute cut planes
            self._cut_planes_origin.SetNumberOfPoints(4)
            self._cut_planes_normal.SetNumberOfTuples(4)

            # top
            self._cut_planes_origin.SetPoint(
                0, self._longitude_bnd[1], self._latitude_bnd[1], 0
            )
            self._cut_planes_normal.SetTuple3(0, 0, -1, 0)
            # bottom
            self._cut_planes_origin.SetPoint(
                1, self._longitude_bnd[0], self._latitude_bnd[0], 0
            )
            self._cut_planes_normal.SetTuple3(1, 0, 1, 0)
            # left
            self._cut_planes_origin.SetPoint(
                2, self._longitude_bnd[0], self._latitude_bnd[0], 0
            )
            self._cut_planes_normal.SetTuple3(2, 1, 0, 0)
            # right
            self._cut_planes_origin.SetPoint(
                3, self._longitude_bnd[1], self._latitude_bnd[1], 0
            )
            self._cut_planes_normal.SetTuple3(3, -1, 0, 0)

        output.ShallowCopy(vtk_mesh)
        return 1



"""
Topo reader for the rivers/lines section
"""

import logging
from pathlib import Path

import h5py
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import (
    vtkCellArray,
    vtkPartitionedDataSet,
    vtkPolyData,
)

from parsli.utils import earth

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


class RiverReader(VTKPythonAlgorithmBase):
    """
    VTK Topo river hdf5 reader
    """

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=0,
            nOutputPorts=1,
            outputType="vtkPartitionedDataSet",
        )
        self._file_name = None
        self._proj_spherical = True
        self._longitude_bnd = [0, 360]
        self._latitude_bnd = [-90, 90]
        self._vertical_scale = 1
        self._max_depth = 0

    @property
    def file_name(self):
        """
        This property captures the file path to read.
        If set using an invalid path, a ValueError is raised.
        """
        return self._file_name

    @file_name.setter
    def file_name(self, path):
        self._file_name = Path(path)
        if not self._file_name.exists():
            msg = f"Invalid file path: {self._file_name.resolve()}"
            raise ValueError(msg)

        self.Modified()

    @property
    def spherical(self):
        """
        This property captures the projection system which can either be spherical or euclidean.
        When set to True (the default), the spherical projection will be used.
        """
        return self._proj_spherical

    @spherical.setter
    def spherical(self, value):
        if self._proj_spherical != value:
            self._proj_spherical = value
            self.Modified()

    @property
    def vertical_scale(self):
        """
        This property captures the vertical scale. The default value is 1.0.
        """
        return self._vertical_scale

    @vertical_scale.setter
    def vertical_scale(self, value):
        if self._vertical_scale != value:
            self._vertical_scale = value
            self.Modified()

    @property
    def maximum_depth(self):
        """
        This property captures the maximum depth based on the mesh read.
        To be accurate, the filter needs to first execute.
        """
        return self._max_depth

    @property
    def longitude_bounds(self):
        """
        This property captures the longitude bounds based on the mesh file.
        """
        return self._longitude_bnd

    @property
    def latitude_bounds(self):
        """
        This property captures the latitude bounds based on the mesh file.
        """
        return self._latitude_bnd

    def _expand_bounds(self, longitude, latitude, depth):
        """Helper to expand longitude and latitude bounds"""
        self._longitude_bnd[0] = min(longitude, self._longitude_bnd[0])
        self._longitude_bnd[1] = max(longitude, self._longitude_bnd[1])

        self._latitude_bnd[0] = min(latitude, self._latitude_bnd[0])
        self._latitude_bnd[1] = max(latitude, self._latitude_bnd[1])

        self._max_depth = max(depth, self._max_depth)

    def RequestData(self, _request, _inInfo, outInfo):
        """VTK Method executed when filter is modified (file_name, spherical)"""
        if self._file_name is None or not self._file_name.exists():
            return 1

        # Reset bounds
        self._longitude_bnd = [360, 0]
        self._latitude_bnd = [90, -90]
        self._max_depth = 0

        # Read file and generate mesh
        output = self.GetOutputData(outInfo, 0)
        all_rivers = vtkPartitionedDataSet()

        # Projection selection
        insert_pt = earth.insert_spherical if self.spherical else earth.insert_euclidian

        with h5py.File(self._file_name, "r") as hdf:
            rivers = hdf["lines"]
            n_rivers = len(rivers)
            all_rivers.SetNumberOfPartitions(n_rivers)
            for idx, name in enumerate(rivers):
                river = rivers[name]
                vtk_mesh = vtkPolyData()
                all_rivers.SetPartition(idx, vtk_mesh)

                # Process known names first
                # - coordinates
                shape = river.shape

                # Generate points
                vtk_points = vtkPoints()
                vtk_points.SetDataTypeToDouble()
                vtk_points.Allocate(shape[0])
                vtk_mesh.points = vtk_points
                vtk_lines = vtkCellArray()
                vtk_mesh.lines = vtk_lines
                vtk_lines.InsertNextCell(shape[0])

                for i in range(shape[0]):
                    lon = river[i][0]
                    lat = river[i][1]
                    elevation = river[i][2]
                    self._expand_bounds(lon, lat, elevation * self._vertical_scale)
                    insert_pt(vtk_points, lon, lat, elevation * self._vertical_scale)
                    vtk_lines.InsertCellPoint(i)

        output.ShallowCopy(all_rivers)
        return 1



"""
Main Parsli viewer
"""

import asyncio
import datetime
import json
import os
import time
from functools import partial
from pathlib import Path

import yaml
from trame.app import TrameApp, asynchronous
from trame.decorators import change, controller
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import html, vtklocal
from trame.widgets import vtk as vtkw
from trame.widgets import vuetify3 as v3
from trame.widgets.trame import MouseTrap
from vtkmodules.vtkFiltersSources import vtkSphereSource
from vtkmodules.vtkIOParallelXML import vtkXMLPartitionedDataSetWriter
from vtkmodules.vtkIOXML import vtkXMLPolyDataWriter

from parsli.io import (
    RiverReader,
    TopoReader,
    VtkCoastLineSource,
    VtkMeshReader,
    VtkSegmentReader,
)
from parsli.utils.core import expend_range, sort_fields, to_precision
from parsli.utils.earth import EARTH_RADIUS
from parsli.utils.source import VtkLatLonBound
from parsli.viewer import css, ui
from parsli.viewer.export import METADATA_STATE_KEYS
from parsli.viewer.vtk import SceneManager


class Viewer(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)

        # Load custom CSS
        self.server.enable_module(css)

        # Add CLI
        self.server.cli.add_argument(
            "--data",
            help="Path of hdf5 file to load",
        )
        self.server.cli.add_argument(
            "--topo",
            help="Path of hdf5 file to load for topo",
        )
        self.server.cli.add_argument(
            "--wasm",
            help="Use local rendering",
            action="store_true",
        )

        # process cli
        args, _ = self.server.cli.parse_known_args()
        self.data_file = str(Path(args.data).resolve()) if args.data else None
        self.local_rendering = args.wasm

        # Handle meta file loading
        if self.data_file:
            input_data = Path(self.data_file)
            input_data = (
                (input_data / "info.yml") if input_data.is_dir() else input_data
            )
            if input_data.exists() and input_data.name == "info.yml":
                meta = yaml.safe_load(input_data.read_text())
                self.data_file = meta.get("info").get("data_path")
                load_meta = partial(self.load_metadata, meta)
                self.ctrl.on_server_ready.add(load_meta)

        # Download sample file if no --data args
        if not args.topo and self.data_file is None:
            from trame.assets.remote import HttpFile

            self.data_file = HttpFile(
                local_path="parsli-sample.hdf5",
                remote_url="https://github.com/brendanjmeade/parsli/raw/refs/heads/main/data/model_0000000927.hdf5",
            ).path

        # Setup app
        self.scene_manager = SceneManager(self.server)
        self._build_ui()

        # Earth core
        pipeline = self.scene_manager.add_geometry(
            "earth_core",
            vtkSphereSource(
                radius=EARTH_RADIUS - 100,
                theta_resolution=60,
                phi_resolution=60,
            ),
        )
        prop = pipeline.get("actor").property
        prop.opacity = 0.85

        # Latitude/Longitude bounding box
        pipeline = self.scene_manager.add_geometry(
            "bbox",
            VtkLatLonBound(),
        )
        bbox_prop = pipeline.get("actor").property
        bbox_prop.line_width = 2
        bbox_prop.color = (0.5, 0.5, 0.5)

        # Quad/Segments meshes
        seg_reader = VtkSegmentReader()
        seg_reader.file_name = self.data_file
        self.state.quad_ui = seg_reader.has_segments
        if seg_reader.has_segments:
            pipeline = self.scene_manager.add_geometry(
                "segment", seg_reader, need_formula=True
            )
            pipeline.get("mapper").SetScalarModeToUseCellFieldData()

        # Surface meshes
        mesh_reader = VtkMeshReader()
        mesh_reader.file_name = self.data_file

        # Extract meshes info for UI
        self.state.fields = sort_fields(mesh_reader.available_fields)
        self.state.time_index = mesh_reader.time_index
        self.state.nb_timesteps = mesh_reader.number_of_timesteps
        self.state.color_by = self.state.fields[0]

        pipeline = self.scene_manager.add_geometry_with_contour(
            "meshes",
            mesh_reader,
            True,
            need_formula=True,
            field_name=self.state.color_by,
        )
        pipeline.get(
            "mapper"
        ).SetScalarModeToUseCellFieldData()  # Bands: Scalars on Cell

        # Coast lines
        self.coast_lines = VtkCoastLineSource()
        self.state.coast_regions = self.coast_lines.available_regions
        self.state.coast_active_regions = []
        pipeline = self.scene_manager.add_geometry("coast", self.coast_lines, True)
        coast_props = pipeline.get("actor").property
        coast_props.line_width = 2
        coast_props.color = (0.5, 0.5, 0.5)

        # Topo
        self.state.topo_ui = bool(args.topo)
        self.terrain = None
        self.rivers = None
        if self.state.topo_ui:
            # terrain
            self.terrain = TopoReader()
            self.terrain.file_name = Path(args.topo).resolve()
            pipeline = self.scene_manager.add_geometry("terrain", self.terrain)
            terrain_props = pipeline.get("actor").property
            terrain_props.color = (0.5, 0.5, 0.5)
            terrain_props.edge_visibility = 1

            # rivers
            self.rivers = RiverReader()
            self.rivers.file_name = Path(args.topo).resolve()
            pipeline = self.scene_manager.add_geometry("rivers", self.rivers, True)
            rivers_props = pipeline.get("actor").property
            rivers_props.color = (0.08, 0.7, 0.96)  # 64B5F6
            rivers_props.line_width = 5

        # setup camera to look at the data
        bounds = self.scene_manager["meshes"].get("actor").bounds
        self.scene_manager.focus_on(bounds)

    @change("color_by")
    def _on_color_by(self, color_by, use_formula, **_):
        pipeline_item = self.scene_manager["meshes"]
        source = pipeline_item.get("source")
        formula = pipeline_item.get("formula")
        mapper_mesh = pipeline_item.get("mapper")

        if color_by is None:
            mapper_mesh.SetScalarVisibility(0)

            if "segment" in self.scene_manager:
                mapper_seg = self.scene_manager["segment"].get("mapper")
                mapper_seg.SetScalarVisibility(0)

            self.ctrl.view_update()
            return

        # Extract data range
        ds = source()

        if use_formula:
            formula.Update()
            ds = formula.GetOutput()

        total_range = None
        for array in ds.cell_data[color_by].Arrays:
            total_range = expend_range(total_range, array.GetRange())

        # prevent min=max
        if total_range[0] == total_range[1]:
            total_range = [
                total_range[0],
                total_range[1] + 1,
            ]

        # Use symmetric range by default
        max_bound = max(abs(total_range[0]), abs(total_range[1]))
        max_bound = to_precision(max_bound, 3)
        self.state.color_min = -max_bound
        self.state.color_max = max_bound

    @change("spherical")
    def _on_projection_change(self, spherical, **_):
        self.state.show_earth_core = spherical

        # Update all meshes with new projection
        for geo_name in ["segment", "meshes", "coast", "bbox", "terrain", "rivers"]:
            if geo_name not in self.scene_manager:
                continue

            pipeline_item = self.scene_manager[geo_name]
            pipeline_item.get("source").spherical = spherical
            actors = pipeline_item.get("actors")

            # In Euclidean mode we need custom z scaling
            scale = (1, 1, 1) if spherical else (1, 1, 0.01)
            for actor in actors:
                actor.scale = scale

        # Update camera based on projection
        if spherical:
            bounds = self.scene_manager["meshes"].get("actor").bounds
            self.scene_manager.camera.focal_point = (0, 0, 0)
            self.scene_manager.camera.view_up = (0, 0, 1)
            self.scene_manager.focus_on(bounds)
        else:
            self.state.interaction_style = "trackball"
            self.scene_manager.camera.focal_point = (0, 0, 0)
            self.scene_manager.camera.position = (0, 0, 1)
            self.scene_manager.camera.view_up = (0, 1, 0)

        self.reset_to_mesh()

    @change("camera")
    def _on_camera(self, camera, **_):
        """To sync camera when doing local rendering with WASM"""
        if camera is None:
            return

        self.ctrl.vtk_update_from_state(camera)

    @change("interaction_style")
    def _on_style_change(self, interaction_style, **_):
        self.scene_manager.update_interaction_style(interaction_style)
        self.ctrl.view_update(push_camera=True)

    @change("subdivide")
    def _on_subdivide(self, subdivide, **_):
        """Change pipeline to smooth or not to smooth surface mesh"""
        # source >> quality >> threshold >> geometry >> cell2point >> refine >> assign >> bands
        pipeline = self.scene_manager["meshes"]

        source = pipeline.get("source")
        formula = pipeline.get("formula")
        assign = pipeline.get("assign")
        refine = pipeline.get("refine")
        cell2point = pipeline.get("cell2point")
        threshold = pipeline.get("threshold")
        geometry = pipeline.get("geometry")

        if subdivide:
            geometry.input_connection = threshold.output_port
            assign.input_connection = refine.output_port

            # debug
            # self.debug_check_quality()
        else:
            geometry.input_connection = (
                formula.output_port if formula else source.output_port
            )
            assign.input_connection = cell2point.output_port

        self.ctrl.view_update()

    @change("screenshot_export_path")
    def _on_export_path(self, screenshot_export_path, **_):
        """Check is directory already exist to prevent export on top of another one"""
        self.state.screenshot_export_path_exits = Path(screenshot_export_path).exists()

    def reset_to_mesh(self):
        """Reset camera to focus on surface mesh bounds"""
        bounds = self.scene_manager["meshes"].get("actor").bounds
        self.scene_manager.reset_camera_to(bounds)
        self.ctrl.view_update(push_camera=True)

    def apply_zoom(self, scale):
        """Zoom in/out base on scale value"""
        self.scene_manager.apply_zoom(scale)
        self.ctrl.view_update(push_camera=True)

    def update_view_up(self, view_up):
        """Snap view-up to provided vector"""
        self.scene_manager.update_view_up(view_up)
        self.ctrl.view_update(push_camera=True)

    async def _export_movie(self):
        """Internal async task to export time series of images"""
        t0 = time.time()
        await asyncio.sleep(0.1)

        # Export path handling
        base_directory = Path(self.state.screenshot_export_path)
        base_directory.mkdir(parents=True)
        self.export_metadata(base_directory)

        print(  # noqa: T201
            "\n----------------------------------------"
            "\nExporting images:"
            f"\n => location: {base_directory.resolve()}"
            f"\n => number of frames: {self.state.nb_timesteps}"
        )

        # Update ScalarBar
        self.scene_manager.update_scalar_bar(
            self.state.color_preset,
            self.state.color_min,
            self.state.color_max,
        )

        # Update Render Window size
        original_size = self.scene_manager.get_size()
        self.scene_manager.set_size(
            self.state.screenshot_width, self.state.screenshot_height
        )

        # Adjust padding which affect scalarbar font size
        if self.state.screenshot_height >= 2160:
            self.scene_manager.scalar_bar.SetTextPad(10)
        elif self.state.screenshot_height >= 1080:
            self.scene_manager.scalar_bar.SetTextPad(5)
        elif self.state.screenshot_height >= 540:
            self.scene_manager.scalar_bar.SetTextPad(2)
        else:
            self.scene_manager.scalar_bar.SetTextPad(1)

        self.scene_manager.show_scalar_bar(True)
        self.scene_manager.render_window.Render()
        self.scene_manager.render_window.SetMultiSamples(4)

        meshes = self.scene_manager["meshes"].get("source")
        segment = None
        if "segment" in self.scene_manager:
            segment = self.scene_manager["segment"].get("source")
        futures = []
        nb_timesteps = self.state.nb_timesteps
        for t_idx in range(nb_timesteps):
            meshes.time_index = t_idx % self.state.nb_timesteps
            if segment:
                segment.time_index = t_idx % self.state.nb_timesteps
            self.scene_manager.set_time(t_idx)
            futures.append(
                self.scene_manager.write_screenshot(base_directory / f"{t_idx:012}")
            )
            progress = int(100 * (t_idx + 1) / nb_timesteps)
            if progress != self.state.export_progress:
                with self.state:
                    self.state.export_progress = progress
                await asyncio.sleep(0.001)

        # Ensure full completion
        for future in futures:
            future.result()

        t2 = time.time()
        print(f" => time: {t2 - t0:.1f}s")  # noqa: T201
        print(f" => fps: {nb_timesteps / (t2 - t0):.1f}")  # noqa: T201

        # Reset size to original
        self.scene_manager.set_size(*original_size)
        self.scene_manager.show_scalar_bar(False)
        self.scene_manager.render_window.SetMultiSamples(1)

        # Done with the export - update the UI
        with self.state:
            self.state.exporting_movie = False
            self.state.export_progress = 100

        print("----------------------------------------")  # noqa: T201

    @controller.set("export_movie")
    def export_movie(self):
        """Called from the UI - trigger background export task for images"""
        self.state.configure_screenshot_export = False
        self.state.export_progress = 0
        self.state.exporting_movie = True
        self.state.screenshot_export_path_exits = True
        asynchronous.create_task(self._export_movie())

    async def _export_data(self):
        """Internal async task to export geometry data"""
        t0 = time.time()
        await asyncio.sleep(0.1)

        # Export path handling
        base_directory = Path(self.state.screenshot_export_path)
        base_directory.mkdir(parents=True)
        self.export_metadata(base_directory)

        print(  # noqa: T201
            "\n----------------------------------------"
            "\nExporting data:"
            f"\n => location: {base_directory.resolve()}"
            f"\n => number of timesteps: {self.state.nb_timesteps}"
        )

        meshes = self.scene_manager["meshes"].get("source")
        segment = self.scene_manager["segment"].get("source")
        bbox = self.scene_manager["bbox"].get("source")
        coast = self.scene_manager["coast"].get("source")

        # For coast lines + surface mesh
        partition_writer = vtkXMLPartitionedDataSetWriter()
        partition_writer.SetInputConnection(coast.output_port)
        partition_writer.SetFileName(str(base_directory / "coast.vtpd"))
        partition_writer.Write()

        # For bbox + segments
        polydata_writer = vtkXMLPolyDataWriter()

        if bbox.valid:
            polydata_writer.SetInputConnection(bbox.output_port)
            polydata_writer.SetFileName(str(base_directory / "bbox.vtp"))
            polydata_writer.Write()

            # Write cutting plane information
            planes_file = base_directory / "planes.json"
            planes_content = []
            planes = bbox.cut_planes
            for i in range(4):
                planes_content.append(
                    {
                        "normal": planes.GetPlane(i).normal,
                        "origin": planes.GetPlane(i).origin,
                    }
                )
            planes_file.write_text(json.dumps(planes_content, indent=2))

        # FIXME - missing topo data...
        # => Don't have the time to add it

        # Time dependent data
        partition_writer.SetInputConnection(meshes.output_port)
        polydata_writer.SetInputConnection(segment.output_port)
        nb_timesteps = self.state.nb_timesteps
        for t_idx in range(nb_timesteps):
            meshes.time_index = t_idx % self.state.nb_timesteps
            segment.time_index = t_idx % self.state.nb_timesteps

            partition_writer.SetFileName(str(base_directory / f"mesh_{t_idx:012}.vtpd"))
            partition_writer.Write()

            polydata_writer.SetFileName(
                str(base_directory / f"segment_{t_idx:012}.vtp")
            )
            polydata_writer.Write()

            progress = int(100 * (t_idx + 1) / nb_timesteps)
            if progress != self.state.export_progress:
                with self.state:
                    self.state.export_progress = progress
                await asyncio.sleep(0.001)

        t2 = time.time()
        print(f" => time: {t2 - t0:.1f}s")  # noqa: T201
        print(f" => timestep per second: {nb_timesteps / (t2 - t0):.1f}")  # noqa: T201

        with self.state:
            self.state.exporting_movie = False
            self.state.export_progress = 100
        print("----------------------------------------")  # noqa: T201

    @controller.set("export_data")
    def export_data(self):
        """Called from the UI - trigger background export task for geometry"""
        self.state.configure_screenshot_export = False
        self.state.export_progress = 0
        self.state.exporting_movie = True
        self.state.screenshot_export_path_exits = True
        asynchronous.create_task(self._export_data())

    def load_metadata_file(self, file):
        """Helper for handling meta data from UI"""
        metadata = yaml.safe_load(file.get("content"))
        self.load_metadata(metadata)

    def load_metadata(self, metadata, **_):
        """Use metadata content to update UI configuration"""
        with self.state:
            self.state.update(metadata.get("state", {}))
            self.scene_manager.update_camera(metadata.get("camera", {}))
            self.ctrl.view_update()
            self.state.configure_screenshot_export = False

    def export_metadata(self, base_path):
        """Write metadata file to disk"""
        metafile = Path(base_path) / "info.yml"
        data = {
            "info": {
                "user": os.environ.get("USER", os.environ.get("USERNAME")),
                "created": f"{datetime.datetime.now()}",
                "data_path": self.data_file,
                "comment": self.state.export_comment,
            },
            "state": {k: self.state[k] for k in METADATA_STATE_KEYS},
            "camera": self.scene_manager.camera_state,
        }
        metafile.write_text(yaml.dump(data))

    def camera_rotate(self, azimuth, elevation):
        """Camera control helper for key binding"""
        self.scene_manager.camera.Azimuth(azimuth)
        self.scene_manager.camera.Elevation(elevation)
        self.ctrl.view_update(push_camera=True)

    def time_move(self, delta):
        """Time control helper for key binding"""
        new_time = self.state.time_index + delta
        new_time = max(min(new_time, self.state.nb_timesteps - 1), 0)
        self.state.time_index = new_time

    async def play_time(self):
        """Loop to move time forward and play animation"""
        while self.state.playing_time:
            if self.state.time_index < self.state.nb_timesteps:
                with self.state:
                    self.state.time_index += 1
                # May need to asjust the wait time
                # => could also be made dynamic
                await asyncio.sleep(0.1)
            else:
                with self.state:
                    self.state.playing_time = False

    @change("playing_time")
    def _on_play_time_change(self, playing_time, **_):
        """Trigger play in background"""
        if playing_time:
            asynchronous.create_task(self.play_time())

    def _build_ui(self):
        self.state.trame__title = "Parsli"
        self.state.setdefault("playing_time", False)
        self.state.setdefault("camera", None)
        with VAppLayout(self.server, full_height=True) as layout:
            # Enable Jupyter usage
            self.ui = layout

            # Keyboard shortcuts
            with MouseTrap(
                # bounds
                boundsSnap=self.ctrl.crop_bound_to_mesh,
                # camera
                cameraReset=self.reset_to_mesh,
                cameraViewUp=(self.update_view_up, "[[0, 0, 1]]"),
                cameraZoomIn=(self.apply_zoom, "[1.1]"),
                cameraZoomOut=(self.apply_zoom, "[0.9]"),
                cameraRotateLeft=(self.camera_rotate, "[-1, 0]"),
                cameraRotateRight=(self.camera_rotate, "[1, 0]"),
                cameraRotateUp=(self.camera_rotate, "[0, 1]"),
                cameraRotateDown=(self.camera_rotate, "[0, -1]"),
                cameraRotateLeftFast=(self.camera_rotate, "[-5, 0]"),
                cameraRotateRightFast=(self.camera_rotate, "[5, 0]"),
                cameraRotateUpFast=(self.camera_rotate, "[0, 5]"),
                cameraRotateDownFast=(self.camera_rotate, "[0, -5]"),
                # time
                timeNextSlow=(self.time_move, "[1]"),
                timeNextMedium=(self.time_move, "[10]"),
                timeNextFast=(self.time_move, "[100]"),
                timePreviousSlow=(self.time_move, "[-1]"),
                timePreviousMedium=(self.time_move, "[-10]"),
                timePreviousFast=(self.time_move, "[-100]"),
                timeFirst="time_index = 0",
                timeLast="time_index = nb_timesteps - 1",
                timeTogglePlay="playing_time = !playing_time",
            ) as mt:
                # view
                mt.bind(["f"], "boundsSnap")
                mt.bind(["r"], "cameraReset")
                mt.bind(["t"], "cameraViewUp")

                # zoom
                mt.bind(["+", "="], "cameraZoomIn")
                mt.bind(["-"], "cameraZoomOut")

                # rotation
                mt.bind(["left"], "cameraRotateLeft")
                mt.bind(["right"], "cameraRotateRight")
                mt.bind(["up"], "cameraRotateUp")
                mt.bind(["down"], "cameraRotateDown")
                mt.bind(["alt+left"], "cameraRotateLeftFast")
                mt.bind(["alt+right"], "cameraRotateRightFast")
                mt.bind(["alt+up"], "cameraRotateUpFast")
                mt.bind(["alt+down"], "cameraRotateDownFast")

                # time
                mt.bind(["space"], "timeTogglePlay")
                mt.bind(["home"], "timeFirst")
                mt.bind(["end"], "timeLast")
                mt.bind(["."], "timeNextSlow")
                mt.bind([">", "alt+."], "timeNextMedium")
                mt.bind(["shift+alt+."], "timeNextFast")
                mt.bind([","], "timePreviousSlow")
                mt.bind(["<", "alt+,"], "timePreviousMedium")
                mt.bind(["shift+alt+,"], "timePreviousFast")

            # Screenshot Export Dialog
            with v3.VDialog(v_model=("configure_screenshot_export", False)):
                with v3.VCard(style="max-width: 50rem;", classes="mx-auto"):
                    with v3.VCardTitle(
                        "Import / Export", classes="d-flex align-center"
                    ):
                        v3.VSpacer()
                        v3.VBtn(
                            icon="mdi-close",
                            flat=True,
                            density="compact",
                            click="configure_screenshot_export = false",
                        )
                    v3.VDivider()
                    with v3.VCardText():
                        with v3.VRow(classes="my-1 align-center"):
                            v3.VTextField(
                                label="Width",
                                v_model=("screenshot_width", 3840),
                                type="number",
                                hide_details=True,
                                density="compact",
                                variant="outlined",
                                hide_spin_buttons=True,
                            )
                            v3.VTextField(
                                label="Height",
                                v_model=("screenshot_height", 2160),
                                type="number",
                                hide_details=True,
                                density="compact",
                                variant="outlined",
                                hide_spin_buttons=True,
                                classes="mx-2",
                            )
                            v3.VBtn(
                                "4K",
                                variant="flat",
                                color="primary",
                                click="screenshot_width=3840; screenshot_height=2160;",
                            )
                            v3.VBtn(
                                "1080p",
                                classes="mx-2",
                                color="secondary",
                                variant="flat",
                                click="screenshot_width=1920; screenshot_height=1080;",
                            )
                            v3.VBtn(
                                "x2",
                                variant="outlined",
                                classes="mx-2",
                                click="screenshot_width=2*screenshot_width; screenshot_height=2*screenshot_height;",
                            )
                            v3.VBtn(
                                "1/2",
                                variant="outlined",
                                click="screenshot_width=0.5*screenshot_width; screenshot_height=0.5*screenshot_height;",
                            )
                        with v3.VRow(classes="my-3"):
                            v3.VTextField(
                                label="Export Path",
                                v_model=(
                                    "screenshot_export_path",
                                    str(Path.cwd().resolve() / "export"),
                                ),
                                density="compact",
                                variant="outlined",
                                error=("screenshot_export_path_exits",),
                                error_messages=(
                                    "screenshot_export_path_exits ? 'Path already exists' : null",
                                ),
                                hide_details=True,
                            )

                        with v3.VRow(classes="my-0"):
                            v3.VTextarea(
                                label="Comments",
                                v_model=("export_comment", ""),
                                density="compact",
                                variant="outlined",
                                rows=5,
                                hide_details=True,
                            )

                        with v3.VRow(classes="mt-2"):
                            with v3.VBtn(
                                "Import",
                                prepend_icon="mdi-import",
                                color="primary",
                                variant="flat",
                                click="utils.get('document').getElementById('import').click()",
                            ):
                                html.Input(
                                    type="file",
                                    id="import",
                                    accept=".yml",
                                    style="display: none",
                                    __events=["change"],
                                    change=(
                                        self.load_metadata_file,
                                        "[$event.target.files[0]]",
                                    ),
                                )
                            v3.VSpacer()

                            v3.VBtn(
                                "Images",
                                prepend_icon="mdi-filmstrip",
                                color="primary",
                                variant="flat",
                                disabled=("screenshot_export_path_exits", False),
                                click=self.ctrl.export_movie,
                                classes="mx-4",
                            )
                            v3.VBtn(
                                "Data",
                                prepend_icon="mdi-database-outline",
                                color="secondary",
                                variant="flat",
                                disabled=("screenshot_export_path_exits", False),
                                click=self.ctrl.export_data,
                            )

            with v3.VContainer(
                fluid=True, classes="fill-height pa-0 ma-0 position-relative"
            ):
                if self.local_rendering:
                    with vtklocal.LocalView(
                        self.scene_manager.render_window,
                        20,
                        camera="camera = $event",
                    ) as view:
                        view.register_vtk_object(self.scene_manager.widget)
                        self.ctrl.view_update = view.update
                        self.ctrl.view_reset_camera = view.reset_camera
                        self.ctrl.vtk_update_from_state = view.vtk_update_from_state
                else:
                    html.Div(
                        v_show="exporting_movie",
                        style="z-index: 1000; position: absolute; width: 100%; height: 100%; top:0; left: 0; background: rgba(0,0,0,0.5); cursor: wait;",
                    )
                    # We could use trame-rca for better rendering performance
                    with vtkw.VtkRemoteView(
                        self.scene_manager.render_window,
                        interactive_ratio=2,
                        still_ratio=2,
                    ) as view:
                        self.ctrl.view_update = view.update
                        self.ctrl.view_reset_camera = view.reset_camera

                # Control panel
                ui.ControlPanel(
                    toggle="show_panel",
                    scene_manager=self.scene_manager,
                    reset_camera=self.ctrl.view_reset_camera,
                    reset_to_mesh=self.reset_to_mesh,
                )

                # 3D View controls
                ui.ViewToolbar(
                    reset_camera=self.ctrl.view_reset_camera,
                    reset_to_mesh=self.reset_to_mesh,
                    apply_zoom=self.apply_zoom,
                    update_view_up=self.update_view_up,
                )

                # ScalarBar
                ui.ScalarBar()
